#------------------------------------------------------------------
# $Copyright: Copyright (c) 2011 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/SunJESMsg/SunJESMsg.pm,v 1.48 2011/10/17 12:37:49 ssrivast Exp $
# $Author: ssrivast $
# $Date: 2011/10/17 12:37:49 $
# $Revision: 1.48 $
#
# Provides:                            SunJESMsg
# Author:                              Anshuman Atre
# Creation Date:                       29 Nov, 2005
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package SunJESMsg;

#---------------------------------
# Include standard PERL packages..
#---------------------------------
use Fcntl;
use Socket;
use POSIX qw ( tmpnam );
use File::Path qw ( mkpath );

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#----------------------------------------------------
# Hard code if you still didn't get it from the env..
#----------------------------------------------------
my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

my $sCmdHares  = $HARES || "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE || "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = $HADECRYPT || "$CLUS_HOME/sbin/${sPrefix}decrypt";

#-------------------------------
# Define list to use in online..
#-------------------------------
my @lsRunningProcs = (); # Global list..

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: SunJESMsg.pm,v 1.48 2011/10/17 12:37:49 ssrivast Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define SunJESMsg subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String ResourceName,
                              IN String LDAPTestPasswd,
                              IN String LDAPTestUser,
                              IN String MonitorProgram,
                              IN String MsgHost,
                              IN RefHash rhMsgServices,
                              IN Integer SecondLevelMonitor,
                              IN String ServerRoot,
                              IN String ServerType );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $sResourceName                 Name of this VCS resource
 $localLDAPTestPasswd           Encrypted password for test LDAP user
 $localLDAPTestUser             Test user on LDAP for monitoring
 $localMonitorProgram           Full path to external monitor program
 $localMsgHost                  Hostname/IPv4 address of virtual host for Messaging Server
 $rhMsgServices                 Services being hosted by this Messaging Server
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 $localServerRoot               Absolute path to server root of Messaging Server
 $localServerType               Role being played by the Messaging Server
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $sResourceName,
       $localLDAPTestPasswd,
       $localLDAPTestUser,
       $localMonitorProgram,
       $localMsgHost,
       $rhMsgServices,
       $localSecondLevelMonitor,
       $localServerRoot,
       $localServerType ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  ResourceName             [%s]
  LDAPTestPasswd           [%s]
  LDAPTestUser             [%s]
  MonitorProgram           [%s]
  MsgHost                  [%s]
  rhMsgServices            [%s]
  SecondLevelMonitor       [%s]
  ServerRoot               [%s]
  ServerType               [%s]",

  $sMyName,          $sEntryPointName,              $sResourceName,
  $localLDAPTestPasswd,  $localLDAPTestUser,  $localMonitorProgram,
  $localMsgHost,     $rhMsgServices,      $localSecondLevelMonitor,
  $localServerRoot, $localServerType );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName LDAPTestPasswd LDAPTestUser MonitorProgram
       MsgServices  MsgHost    MsgServices    SecondLevelMonitor
       ServerRoot  ServerType );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( MsgHost ServerRoot ServerType );
  my @lsOfflineArgs = qw ( ServerRoot ServerType );
  my @lsMonitorArgs = qw ( ServerRoot ServerType );
  my @lsCleanArgs   = qw ( ServerRoot ServerType );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #--------------------------------------
  # Validate ServerType and MsgServices..
  #--------------------------------------
  my %hsServCopy = %$rhMsgServices; # Backup..
  unless ( $iReturnCode = SunJESMsg::GetMsgServDetails ( $localServerRoot,
                                                         \%hsServCopy,
                                                         $localServerType ) ) {
    #------------------------------------
    # Display failed MsgServices values..
    #------------------------------------
    if ( scalar keys %hsServCopy ) {
      &logError ( 20003,
      "<%s> MsgServices [%s] could not be validated for ServerType [%s]",
      $sMyName, join ( ' ', keys %hsServCopy ), $localServerType );
    }
  }

  #------------------------------
  # Return, if monitor or clean..
  #------------------------------
  if ( $sEntryPointName eq 'monitor' ||
       $sEntryPointName eq 'clean' ) {
    return $iReturnCode;
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------

  #-------------------------
  # Server root must exist..
  #-------------------------
  if ( ! -d $localServerRoot ) {
    &logError ( 20004,
    "<%s> ServerRoot directory [%s] does not exist",
    $sMyName, $localServerRoot );
    $iReturnCode = 0; # Failure..
  }

  if ( $sEntryPointName eq 'online' ) {
    #--------------------
    # Validate hostname..
    #--------------------
    if ( ! inet_aton ( $localMsgHost ) ) {
      &logError ( 20005, "<%s> Could not resolve MsgHost [%s]",
      $sMyName, $localMsgHost );
      $iReturnCode = 0; # Failure..
    }

    #--------------------------------------
    # MonitorProgram has to be executable..
    #--------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ &&     # Defined..
         ! -x $1 ) {                            # And not executable..
      &logError ( 20006,
      "<%s> MonitorProgram [%s] not executable", $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }
    
    #-----------------
    # LDAPTestPasswd..
    #-----------------
    if ( $localLDAPTestUser =~ /\S/ &&    # Defined..
         $localLDAPTestPasswd !~ /\S/ ) { # And not defined..
      &logError ( 20007,
      "<%s> LDAPTestPasswd cannot be NULL [%s] if LDAPTestUser [%s] is defined",
      $sMyName, $localLDAPTestPasswd, $localLDAPTestUser );
      $iReturnCode = 0; # Failure..
    }
  }

  $iReturnCode ||
  &logWarn ( 20008,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN RefHash rhMsgServices,
                         IN String ServerRoot,
                         IN String ServerType );

 Description : Start logic needed for online entry pt

 Parameters :

 $rhMsgServices         Services being hosted by this Messaging Server
 $sServerRoot           Absolute path to server root of Messaging Server
 $sServerType           Role being played by the Messaging Server
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rhMsgServices,
       $sServerRoot,
       $sServerType ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20009, "Subroutine <%s> called with:
  rhMsgServices            [%s]
  ServerRoot               [%s]
  ServerType               [%s]",

  $sMyName, $rhMsgServices, $sServerRoot, $sServerType );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------------
  # Initialise local variables..
  #-----------------------------
  my ( $sCmd, $sCmdArgs, $sService, $iTillWhen, $iExitCode,
       $iTimeout, @lsOldLibPath, @lsCmdOut, @lsServices, @liPids,
       %hsServices ) = ();

  my $iSleepTime = 10; # Secs..

  #--------------------------------
  # Kill any process if remaining..
  #--------------------------------
  if ( SunJESMsg::KillMsgProcs ( $sServerRoot,
                                 $sServerType,
                                 $rhMsgServices ) ) {
    &logTrace ( 20010,
    "<%s> Confirmed no persistent processes exist before startup",
    $sMyName );
  }
  else {
    &logError ( 20011,
    "<%s> Persistent processes exist for Messaging Server", $sMyName );
  }

  #----------------------------
  # Set environment variables..
  #----------------------------
  if ( SunJESMsg::SetMsgEnv ( $sServerRoot, \@lsOldLibPath ) ) {
    &logTrace ( 20012,
    "<%s> Set up environment for running Msg server command",
    $sMyName );
  }
  else {
    &logError ( 20013,
    "<%s> Failed to set up environment for running Msg server command",
    $sMyName );
    return 0;
  }

  #-----------------
  # Form start cmd..
  #-----------------
  $sCmd = "$sServerRoot/lib/msstart";

  if ( ! -e $sCmd ) {
    &logError ( 20014, "<%s> Start command [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20015, "<%s> Start command [%s] not executable",
    $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  else {
    &logTrace ( 20016, "<%s> Located start command [%s] for execution",
    $sMyName, $sCmd );
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #----------------------
  # Form start cmd args..
  #----------------------
  if ( $sServerType eq 'MSG' ) {
    @lsServices = keys %$rhMsgServices; # Backup..

    foreach $sService ( @lsServices ) {
      $sService =~ s/^(imap|http|pop)s/$1/; # Remove SSL factor..
      $sService =~ s/^core$/sched store/;   # Re-create core..
      ++ $hsServices { $sService };         # Remove duplicates..
    }

    $sCmdArgs = join ( ' ',  keys %hsServices );
    delete @hsServices { keys %hsServices }; # Free..
  }
  else { # MMP..
    $sCmdArgs = 'mmp';
  }

  #--------------------
  # Get time bearings..
  #--------------------
  $iTimeout = SunJESMsg::GetEngineTimeout ( 1 ); # With margin..
  $iTillWhen = $^T + $iTimeout - time;

  &logTrace ( 20017,
  "<%s> Starting server as [%s] with Cmd [%s] Args [%s] within [%s] sec",
  $sMyName, $sServerType, $sCmd, $sCmdArgs, $iTillWhen );

  #---------------
  # Fire command..
  #---------------
  if ( $iReturnCode = SunJESMsg::ExecLocalCmd ( $sCmd,
                                                $sCmdArgs,
                                                \$iExitCode,
                                                $iTillWhen,
                                                \@lsCmdOut ) ) {
    &logTrace ( 20018,
    "<%s> Command [%s] executed successfully with exit code [%s]",
    $sMyName, $sCmd, $iExitCode );

    @lsCmdOut = (); # Free..

    #--------------------
    # Delay entry point..
    #--------------------

    while ( ( $iTillWhen = $^T + $iTimeout - time )  > 10 ) {

      if ( VCSDef_MonitorExit::ONLINE ==
           SunJESMsg::GetProcState ( $sServerRoot,
                                     $sServerType,
                                     \@liPids,
                                     $rhMsgServices,
                                     1 ) ) { # Try fast FLM..
        &logInfo ( 20019,
        "<%s> Confirmed startup of Msg Server as [%s] with PIDs [%s]",
        $sMyName, $sServerType, join ( ' ', @liPids ) );
        $iReturnCode = 1; # Success..
        last; # Get out..
      }    

      &logTrace ( 20020, "<%s> Could not confirm startup.
      Delaying exit with [%s] sec in hand", $sMyName, $iTillWhen );

      $iReturnCode = 0;  # Failure.
      sleep $iSleepTime; # Pause before re-trying..
    }
  }
  else {
    &logError ( 20021,
    "<%s> Command [%s] execution failed with exit code [%s]. Cmd output [%s]",
    $sMyName, $sCmd, $iExitCode, join ( chr 012, @lsCmdOut ) );
    $iReturnCode = 0; # Failure..
  } 

  #------------------------
  # Restore env variables..
  #------------------------
  if ( scalar @lsOldLibPath ) {
    SunJESMsg::SetMsgEnv ( $sServerRoot,
                           \@lsOldLibPath );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN RefHash rhMsgServices,
                        IN String ServerRoot,
                        IN String ServerType );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $rhMsgServices         Services being hosted by this Messaging Server
 $sServerRoot           Absolute path to server root of Messaging Server
 $sServerType           Role being played by the Messaging Server
 

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rhMsgServices,
       $sServerRoot,
       $sServerType ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20022, "Subroutine <%s> called with:
  rhMsgServices            [%s]
  ServerRoot               [%s]
  ServerType               [%s]",

  $sMyName, $rhMsgServices, $sServerRoot, $sServerType );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-----------------------------
  # Initialise local variables..
  #-----------------------------
  my ( $sCmd, $sCmdArgs, $sService, $iTillWhen, $iTimeout, $iExitCode,
       @lsOldLibPath, @lsCmdOut, @lsServices, @liPids, %hsServices ) = ();

  my $iSleepTime = 10; # Secs..

  #----------------------------
  # Set environment variables..
  #----------------------------
  if ( SunJESMsg::SetMsgEnv ( $sServerRoot, \@lsOldLibPath ) ) {
    &logTrace ( 20023,
    "<%s> Set up environment for running Msg server command",
    $sMyName );
  }
  else {
    &logError ( 20024,
    "<%s> Failed to set up environment for running Msg server command",
    $sMyName );
  }

  #----------------------------
  # Form stop cmd in any case..
  #----------------------------
  $sCmd = "$sServerRoot/lib/msstart";
  $sCmdArgs = "-k ";

  if ( ! -e $sCmd ) {
    &logError ( 20025, "<%s> Stop command [%s] does not exist",
    $sMyName, $sCmd );
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20026, "<%s> Stop command [%s] not executable",
    $sMyName, $sCmd );
  }
  else {
    &logTrace ( 20027, "<%s> Located stop command [%s] for execution",
    $sMyName, $sCmd );

    #----------------------
    # Form stop cmd args..
    #----------------------
    if ( $sServerType eq 'MSG' ) {
      @lsServices = keys %$rhMsgServices; # Backup..
  
      foreach $sService ( @lsServices ) {
        $sService =~ s/^(imap|http|pop)s/$1/;       # Remove SSL factor..
        $sService =~ s/^core$/sched store watcher/; # Re-create core..
        ++ $hsServices { $sService };               # Remove duplicates..
      }
  
      $sCmdArgs .= join ( ' ',  keys %hsServices );
      delete @hsServices { keys %hsServices }; # Free..
    }
    else { # MMP..
      $sCmdArgs .= 'mmp watcher';
    }

    #--------------------
    # Get time bearings..
    #--------------------
    $iTimeout = SunJESMsg::GetEngineTimeout ( 1 ); # With margin..
    $iTillWhen = $^T + $iTimeout - time;
  
    &logTrace ( 20028,
    "<%s> Stopping server as [%s] with Cmd [%s] Args [%s] within [%s] sec",
    $sMyName, $sServerType, $sCmd, $sCmdArgs, $iTillWhen );
  
    #---------------
    # Fire command..
    #---------------
    if ( $iReturnCode = SunJESMsg::ExecLocalCmd ( $sCmd,
                                                  $sCmdArgs,
                                                  \$iExitCode,
                                                  $iTillWhen,
                                                  \@lsCmdOut ) ) {
      &logTrace ( 20029,
      "<%s> Command [%s] executed successfully with exit code [%s]",
      $sMyName, $sCmd, $iExitCode );
  
      @lsCmdOut = (); # Free..
  
      #--------------------
      # Delay entry point..
      #--------------------
      while ( ( $iTillWhen = $^T + $iTimeout - time )  > 10 ) {
  
        @liPids = (); # Reset..
        SunJESMsg::GetProcState ( $sServerRoot,
                                  $sServerType,
                                  \@liPids,
                                  $rhMsgServices,
                                  0 ); # Use regular FLM..

        #-----------------------------
        # Persistant processes found..
        #-----------------------------
        if ( scalar @liPids ) {
          &logTrace ( 20030,
          "<%s> Could not confirm shutdown. [%s] processes remain.
          Delaying kill with [%s] sec in hand. Retrying after [%s] sec",
          $sMyName, scalar @liPids, $iTillWhen, $iSleepTime );
  
          sleep $iSleepTime; # Pause before re-trying..
        }
        else {
          &logInfo ( 20031,
          "<%s> Confirmed shutdown of Msg Server as [%s]",
          $sMyName, $sServerType );
          last; # Get out..
        }
      }
    }
    else {
      &logError ( 20032,
      "<%s> Command [%s] execution failed with exit code [%s]. Cmd output [%s]",
      $sMyName, $sCmd, $iExitCode, join ( chr 012, @lsCmdOut ) );
    }
  }

  #--------------------------------
  # Kill any process if remaining..
  #--------------------------------
  if ( SunJESMsg::KillMsgProcs ( $sServerRoot,
                                 $sServerType,
                                 $rhMsgServices ) ) {
    &logTrace ( 20033,
    "<%s> Confirmed no pending processes exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20034,
    "<%s> Pending processes exist for Messaging Server", $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #------------------------
  # Restore env variables..
  #------------------------
  if ( scalar @lsOldLibPath ) {
    SunJESMsg::SetMsgEnv ( $sServerRoot,
                           \@lsOldLibPath );
  }
  
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN RefHash rhMsgServices,
                        IN String ServerRoot,
                        IN String ServerType,
                        IN Integer CleanCode );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $rhMsgService          Services being hosted by this Messaging Server
 $sServerRoot           Absolute path to server root of Messaging Server
 $sServerType           Role being played by the Messaging Server
 $iCleanCode            Why clean was called
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rhMsgServices,
       $sServerRoot,
       $sServerType,
       $iCleanCode ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanCode } ||         # Reason..
                    $iCleanCode ) =~ s/_/ /g;

  &logTrace ( 20035, "Subroutine <%s> called with:
  rhMsgServices            [%s]
  ServerRoot               [%s]
  ServerType               [%s]
  Clean reason             [%s]",

  $sMyName, $rhMsgServices, $sServerRoot, $sServerType,
  $sCleanReason );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------------
  # Initialise local variables..
  #-----------------------------
  my ( $sCmd, $sCmdArgs, $sService, $iTillWhen, $iExitCode,
       @lsOldLibPath, @lsCmdOut, @lsServices, @liPids,
       %hsServices ) = ();

  #-------------------
  # Try to be gentle..
  #-------------------
  unless ( $hsCleanReasons { $iCleanCode } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanCode } eq 'Offline_was_ineffective' ) {
    #----------------------------
    # Set environment variables..
    #----------------------------
    if ( SunJESMsg::SetMsgEnv ( $sServerRoot, \@lsOldLibPath ) ) {
      &logTrace ( 20036,
      "<%s> Set up environment for running Msg server command",
      $sMyName );
    }
    else {
      &logError ( 20037,
      "<%s> Failed to set up environment for running Msg server command",
      $sMyName );
    }
  
    #----------------------------
    # Form stop cmd in any case..
    #----------------------------
    $sCmd = "$sServerRoot/lib/msstart";
    $sCmdArgs = "-k ";
  
    if ( ! -e $sCmd ) {
      &logError ( 20038, "<%s> Stop command [%s] does not exist",
      $sMyName, $sCmd );
    }
    elsif ( ! -x $sCmd ) {
      &logError ( 20039, "<%s> Stop command [%s] not executable",
      $sMyName, $sCmd );
    }
    else {
      &logTrace ( 20040, "<%s> Located stop command [%s] for execution",
      $sMyName, $sCmd );
  
      #----------------------
      # Form stop cmd args..
      #----------------------
      if ( $sServerType eq 'MSG' ) {
        @lsServices = keys %$rhMsgServices; # Backup..
    
        foreach $sService ( @lsServices ) {
          $sService =~ s/^(imap|http)s/$1/;   # Remove SSL factor..
          $sService =~ s/^core$/sched store watcher/; # Re-create core..
          ++ $hsServices { $sService };       # Remove duplicates..
        }
    
        $sCmdArgs .= join ( ' ',  keys %hsServices );
        delete @hsServices { keys %hsServices }; # Free..
      }
      else { # MMP..
        $sCmdArgs .= 'mmp watcher';
      }
  
      #--------------------
      # Get time bearings..
      #--------------------
      $iTillWhen = $^T +
                   SunJESMsg::GetEngineTimeout ( 1 ) - time; # With margin..
    
      &logTrace ( 20041,
      "<%s> Stopping server as [%s] with Cmd [%s] Args [%s] within [%s] sec",
      $sMyName, $sServerType, $sCmd, $sCmdArgs, $iTillWhen );
    
      #---------------
      # Fire command..
      #---------------
      if ( SunJESMsg::ExecLocalCmd ( $sCmd,
                                     $sCmdArgs,
                                     \$iExitCode,
                                     $iTillWhen,
                                     \@lsCmdOut ) ) {
        &logInfo ( 20042,
        "<%s> Command [%s] executed successfully with exit code [%s]",
        $sMyName, $sCmd, $iExitCode );
      }
      elsif ( ! $iExitCode ) {
        &logError ( 20043,
        "<%s> Command [%s] with args [%s] timed out",
        $sMyName, $sCmd, $sCmdArgs );
      }
      else {
        &logError ( 20044,
        "<%s> Command [%s] with args [%s] failed with exit code [%s].
        Cmd output [%s]",
        $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
      }
    }
  
    #------------------------
    # Restore env variables..
    #------------------------
    if ( scalar @lsOldLibPath ) {
      SunJESMsg::SetMsgEnv ( $sServerRoot,
                             \@lsOldLibPath );
    }
  }

  #--------------------------------
  # Kill any process if remaining..
  #--------------------------------
  if ( SunJESMsg::KillMsgProcs ( $sServerRoot,
                                 $sServerType,
                                 $rhMsgServices ) ) {
    &logTrace ( 20045,
    "<%s> Confirmed no pending processes exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20046,
    "<%s> Pending processes exist for Messaging Server", $sMyName );
    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN RefHash rhMsgServices,
                                   IN String MsgHost,
                                   IN String ServerRoot,
                                   IN String ServerType );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sMsgServices          Services being hosted by this Messaging Server
 $sMsgHost              Hostname/IPv4 address of virtual host for Messaging Server
 $sServerRoot           Absolute path to server root of Messaging Server
 $sServerType           Role being played by the Messaging Server

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  PARTIAL                                  VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rhMsgServices,
       $sMsgHost,
       $sServerRoot,
       $sServerType ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20047, "Subroutine <%s> called:
  rhMsgServices            [%s]
  MsgHost                  [%s]
  ServerRoot               [%s]
  ServerType               [%s]",

  $sMyName, $rhMsgServices, $sMsgHost, $sServerRoot, $sServerType, );

  my @liDummyPids = ();
  my $iReturnCode = SunJESMsg::GetProcState ( $sServerRoot,
                                              $sServerType,
                                              \@liDummyPids,
                                              $rhMsgServices,
                                              1 ); # Try fast FLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20049,
    "<%s> Found all required processes", $sMyName );
  }
  else {
    &logTrace ( 20050,
    "<%s> Did not find all required processes, returning [%s]",
    $sMyName, $iReturnCode );
  }

  @lsRunningProcs = @liDummyPids; # Populate global..

  &logTrace ( 20051,
  "<%s> Filling global array with [%s] elements and Returning [%s]",
  $sMyName, scalar @lsRunningProcs, $iReturnCode );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( IN String ServerRoot,
                               IN String ServerType,
                               OUT ListReference OnlinePids,
                               IN RefHash MsgServices,
                               In Boolean UseFFLM );

 Description : Performs the actual first level monitoring

 Parameters :

 $sServerRoot      Absolute path to ServerRoot of Admin Server
 $sServerType      Role being played by the Messaging Server
 $rliOnlinePids    PIDs of relevant online processes
 $rhMsgServices    Services being hosted by this Messaging Server
 $bUseFFLM         Flag to indicate if fast FLM is to be attempted

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  PARTIAL                                  VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerRoot,
       $sServerType,
       $rliOnlinePids,
       $rhMsgServices,
       $bUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20052, "Subroutine <%s> called:
  ServerRoot      [%s]
  ServerType      [%s]
  rliOnlinePids   [%s]
  rhMsgServices   [%s]
  UseFFLM         [%s]",

  $sMyName,
  $sServerRoot, $sServerType, $rliOnlinePids,
  $rhMsgServices, $bUseFFLM );

  #----------------------------------------
  # Set Return Code Optimistically to 110..
  #----------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sKey, $sTemp, $sPidFile, $sProc, $sPattern, $sPrKey,
       $sEntryPoint, $sLang, $iFound, $iProcessCount, $iUseRFLM, $iPid,
       $iRedo, $iSleepTime, $rhlProc, @lsUpdates, @hProcListHash,
       @lsValues, %hsServCopy, %hiProcPids ) = ();

  @$rliOnlinePids = (); # Reset..
  ( $sEntryPoint = $0 ) =~ s#.*/##;        # Basename..

  #--------------------------------
  # First backup Msg Services hash!
  #--------------------------------
  $iUseRFLM = scalar ( keys %$rhMsgServices ); # Set for regular FLM..
  $iRedo = 2;                                  # Attempt max twice..
  $iSleepTime = 10;                            # Secs..

  while ( $iRedo ) {
    %hsServCopy = %$rhMsgServices;        # Save..

    #--------------------------------
    # Get PID file and process name..
    #--------------------------------
    if ( SunJESMsg::GetMsgServDetails ( $sServerRoot,
                                        \%hsServCopy,
                                        $sServerType ) ) {

      #-------------------------------------------------------
      # Msg server configured for restart of failed services..
      #-------------------------------------------------------
      if ( $sEntryPoint eq 'monitor' ) {
        -- $iRedo;
        &logTrace ( 20053,
        "<%s> Assuming Watcher enabled for restart. Setting redo count to [%s]",
        $sMyName, $iRedo );
      }
      else {
        $iRedo = 0; # Reset.. 
        &logTrace ( 20054,
        "<%s> Ignoring Watcher restart capability. Setting redo count to [%s]",
        $sMyName, $iRedo );
      }
  
      #--------------------------------------------
      # Use only one PID file and process for MMP..
      #--------------------------------------------
      if ( $sServerType eq 'MMP' ) {
        $hsServCopy { AService } =               # Define AService..
        $hsServCopy { ( keys %hsServCopy )[0] }; # As a copy..

        grep { $_ ne 'AService' &&               # If not AService..
               delete $hsServCopy { $_ }         # Then delete it..
        } keys %hsServCopy;                      # From the hash..

        $iUseRFLM = 1;
      }
  
      if ( $bUseFFLM ) {
        $iUseRFLM = 0; # Hopefully skip..
        #----------------------
        # Form display string..
        #----------------------
        foreach $sKey ( keys %hsServCopy ) {
          @lsUpdates = (); # Reset..
  
          @lsValues = @{ $hsServCopy { $sKey } }; # Get values..
          $sTemp .= "\nService      [$sKey]\n";
    
          while ( ( $sPidFile, $sProc ) =
                    splice ( @lsValues, 0, 2 ) ) {
    
            #---------------
            # Run fast FLM..
            #---------------
            $iFound = SunJESMsg::FFLMStatus ( $sPidFile,
                                              $sProc,
                                              \$iPid );
  
            #---------------------
            # Handle exit status..
            #---------------------
            if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
                push ( @$rliOnlinePids, $iPid ); # Store..
            }
            elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
                ++ $iUseRFLM;                    # Regular FLM on service..
            }
            else { # VCSDef_MonitorExit::UNKNOWN:
                push ( @$rliOnlinePids, $iPid ); # Store..
            }
  
            #----------------
            # Update status..
            #----------------
            push ( @lsUpdates, $sPidFile, $sProc, $iFound );
  
            $sTemp .= "PID File     [$sPidFile]\n";
            $sTemp .= "Srch Process [$sProc]\n";
            $sTemp .= "FFLM Status  [$iFound]\n\n";
          }
  
          @{ $hsServCopy { $sKey } } = @lsUpdates; # Update..
        }
  
        &logTrace ( 20055,
        "<%s> ServerType [%s] MsgService details ---%s\n",
         $sMyName, $sServerType, $sTemp );
    
        ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
      }
      else { # Use regular FLM..
        #-----------------------------------------
        # Mark status of each service as offline..
        #-----------------------------------------
        $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
    
        foreach $sKey ( keys %hsServCopy ) {
          @lsUpdates = (); # Reset..
  
          @lsValues = @{ $hsServCopy { $sKey } }; # Get values..
          while ( scalar @lsValues ) {
            push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
          }
  
          @{ $hsServCopy { $sKey } } = @lsUpdates; # Update..
        }
      }
    }
    else {
      &logError ( 20056,
      "<%s> MsgServices [%s] could not be validated for ServerType [%s]",
      $sMyName, join ( ' ', keys %hsServCopy ), $sServerType );
      return VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
  
    #--------------------------------------------------------
    # We need to run the regular (slow :-o) ps(1m) command..
    #--------------------------------------------------------
    if ( $iUseRFLM ) {
      &logTrace ( 20057,
      "<%s> Running regular first level monitor", $sMyName );

      #----------------------------------
      # Set LANG for better performance..
      #----------------------------------
      $sLang = $ENV { LANG }; # Backup..
      $ENV { LANG } = 'C';    # Now set..

      #-----------------------------------
      # Call processes and store in hash..
      #-----------------------------------
      $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

      #----------
      # Restore..
      #----------
      defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..
    
      #----------------------
      # Check if we got any..
      #----------------------
      if ( 0 == $iProcessCount ) { # Oops!
        &logError ( 20058,
        "<%s> Call to get process list did not return any processes",
        $sMyName );
        return VCSDef_MonitorExit::OFFLINE; # Offline..
      }
    
      #-----------------------
      # Search for processes..
      #-----------------------
      foreach $rhlProc ( @hProcListHash ) {
        #--------------------------------------------------------------
        # Each $rhlProc is a ref. to a hash with the following keys..
        # "UID"     =>  <user id>
        # "PID"     =>  <process id>
        # "PPID"    =>  <parent process id>
        # "COMMAND" =>  <command string>
        #--------------------------------------------------------------

        next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
        next if $$rhlProc { PID } == $$;       # Skip self..
        $iPid = $$rhlProc { PID };             # Assign..
    
        &logTrace ( 20059, "<%s> Searching PID [%s] PPID [%s] CMD [%s]",
        $sMyName, $$rhlProc { PID }, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
  
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        foreach $sKey ( keys %hsServCopy ) {
          @lsUpdates = (); # Reset..
          @lsValues = @{ $hsServCopy { $sKey } }; # Get values..
  
          #--------------------------------------------------
          # Should have details per three elements of array..
          # <PIDFILE> <PROCESS> <STATE>
          #--------------------------------------------------
          while ( ( $sPidFile, $sProc, $iFound ) =
                    splice ( @lsValues, 0, 3 ) ) {
  
            #--------------------------------------
            # Verify if process is really offline..
            #--------------------------------------
            if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
              if ( $$rhlProc { COMMAND } =~ /^$sProc\b/ ) { # But now found..
                &logTrace ( 20060,
                "<%s> Found PID [%s] CMD [%s] online in regular FLM with filter [%s]",
                $sMyName, $iPid, $$rhlProc { COMMAND }, $sProc );

                ( $sPrKey ) = $sPidFile =~ m#pidfile\.(\S+)$#;
                $hiProcPids { $sPrKey } = $iPid;  # Store..

                $iFound = VCSDef_MonitorExit::ONLINE;       # Update..
                push ( @$rliOnlinePids, $$rhlProc { PID } );# Store..
              }
            }
  
            #----------------
            # Update status..
            #----------------
            push ( @lsUpdates, $sPidFile, $sProc, $iFound );
          }
  
          @{ $hsServCopy { $sKey } } = @lsUpdates; # Update..
        }
      }
    }
  
    #--------------------------------------
    # Print final status for each service..
    #--------------------------------------
    my ( $iOnline, $iOffline, $iUnknown ) = ();
    $sTemp = ''; # Ensure..
  
    foreach $sKey ( keys %hsServCopy ) {
      @lsValues = @{ $hsServCopy { $sKey } }; # Get values..
      $sTemp .= "\nService      [$sKey]\n";
  
      while ( ( $sPidFile, $sProc, $iFound ) =
                splice ( @lsValues, 0, 3 ) ) {
        #-----------------------------------
        # Create PID files for use in FFLM..
        #-----------------------------------
        ( $sPrKey ) = $sPidFile =~ m#pidfile\.(\S+)$#;

        if ( $iUseRFLM                                 &&  # In RFLM..
             $hiProcPids { $sPrKey }                   &&  # PID populated..
             ( $sEntryPoint =~ /^(monitor|online)$/o ) ) { # In monitor or online..

          &logTrace ( 20061,
          "<%s> Creating PID file [%s] with PID [%s]",
          $sMyName, $sPidFile, $hiProcPids { $sPrKey } );
  
          #---------
          # Create..
          #---------
          if ( SunJESMsg::CreatePidFile ( $sPidFile,
                                          $hiProcPids { $sPrKey } ) ) {
            &logTrace ( 20062,
            "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
          }
          else {
            &logError ( 20063,
            "<%s> Could not create PID file [%s] successfully",
            $sMyName, $sPidFile );
          }
        }
    
        $sTemp .= "Srch Process [$sProc]\n";
        $sTemp .= "FLM Status   [$iFound]\n\n";
  
        #------------------------ 
        # Handle process status..
        #------------------------ 
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          $iOnline = $iFound;
        }
        elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Skip watcher..
          $iOffline = $iFound unless ( $sPidFile =~ /pidfile.watcher$/ );
        }
        else { # VCSDef_MonitorExit::UNKNOWN
          $iUnknown = $iFound;
        }
      }
  
      $sTemp .= "--------------------";
    }
  
    #-------------------
    # Set final status..
    #-------------------
    $iReturnCode = $iOnline && $iOffline?VCSDef_MonitorExit::PARTIAL_ONLINE:
                               $iOffline || $iUnknown || $iOnline;

    &logTrace ( 20064,
    "<%s> ServerType [%s] MsgService details ---\n%s
    Return code is [%s]",
    $sMyName, $sServerType, $sTemp, $iReturnCode );
  
    $sTemp = undef; # Release..

    #--------------------------------
    # Get out if we are not offline..
    #--------------------------------
    if ( VCSDef_MonitorExit::OFFLINE != $iReturnCode ) {
      $iRedo = 0;
    }
    elsif ( $iRedo ) { # Pause before re-attempting..
      &logTrace ( 20065,
      "<%s> Pausing [%s] secs before re-attempting",
      $sMyName, $iSleepTime );

      sleep $iSleepTime; # Take a wink..
    }
  }

  return $iReturnCode; # Return status..
}

=pod

 -------------------------------------------------------------------------
 Function : bool FFLMStatus ( IN String PidFile,
                              IN String Proc,
                              OUT RefInt Pid );

 Description : Checks status of Process listed in PidFile, if
               listed as 'ready'

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20066, "Subroutine <%s> called:
  PidFile         [%s]
  Proc            [%s]
  riPid           [%s]",

  $sMyName, $sPidFile, $sProc, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, @lsProcessHash, $rhlProc, $iExitCode, @lsFile ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20067, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20068, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20069, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;

  if ( open ( PIDFILE, $sPidFile ) ) {
  
    @lsFile = grep { s/\n// || 1 } <PIDFILE>; # Read in file..

    close ( PIDFILE ) ||                      # Close it..
    &logWarn ( 20070,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );

    $lsFile [0] = '' unless defined $lsFile [0]; # Avoid undefs..

    #---------------------------
    # Check if PID file status..
    #---------------------------
    if ( $lsFile [0] =~ /\d+/ &&     # If the valid number..
         kill ( 0, $lsFile [0] ) ) { # is a PID that's alive..
      $iPid = $lsFile [0];           # Assign..

      &logTrace ( 20071,
      "<%s> PID [%s] from PID file [%s] exists",
      $sMyName, $iPid, $sPidFile );

      #-----------------------
      # Search in proc table..
      #-----------------------
      if ( SunJESMsg::GetPidDetails ( $iPid,
                                      \@lsProcessHash ) ) {
        #--------------
        # Got details..
        #--------------
        foreach $rhlProc ( @lsProcessHash ) {
          &logTrace ( 20072,
          "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
          $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
          $$rhlProc { COMMAND } );
  
          #--------------
          # Now compare..
          #--------------
          if ( $$rhlProc { COMMAND } =~ /^$sProc\b/ ) { # Filter match..
            &logTrace ( 20073,
            "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
            $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
            $$riPid = $iPid; # Store..
          }
          else {
            &logTrace ( 20074,
            "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
            $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
            $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
          }
        }
      }
      else {
        &logTrace ( 20075,
        "<%s> Could not get details for PID [%s]",
        $sMyName, $iPid );
        $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
      }
    }
    else { # Corrupt PID file..
      &logTrace ( 20076, "<%s> PID [%s] from PID File [%s] does not exist",
      $sMyName, $iPid, $sPidFile );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }

  return $iReturnCode; # Return status..
}

=pod

 -------------------------------------------------------------------------
 Function : int GetMsgServDetails ( IN String ServerRoot,
                                    INOUT RHS MsgServices,
                                    IN String ServerType );

 Description : For a valid Msg Server service, gives the corresponding
               PID file and process name, if successful.

               Also validates against supported services. Ports have
               to be greater than 0.

               Also validates if ServerType is one of MSG or MMP, and
               if MMP, MsgServices is only one of imap, imaps or pop.

 Valid services:

               snmp, sms, ens, pop, http,
               https, imap, imaps, smtp

               OR

               core

               Else updates hash with services in error.
 E.g:

 %hsMsgServices = qw ( http 80 imap 143 pop 110 smtp 25 sms 443 );
 %hsMsgServices = qw ( core 49994 ); # Port of watcher..

 Parameters  :

 $sServerRoot       Absolute path to ServerRoot of Msg Server
 $rhsMsgService     Ref to hash of MsgServices
 $sServerType       Role being played by the Messaging Server

 Returns:  boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetMsgServDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerRoot,
       $rhsMsgServices,
       $sServerType ) = @_;
 
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20077, "Subroutine <%s> called:
  ServerRoot               [%s]
  rhsMsgServices           [%s]
  ServerType               [%s]",

  $sMyName, $sServerRoot, $rhsMsgServices,
  $sServerType );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #----------------------
  # Set local variables..
  #----------------------
  my @lsServices = keys %$rhsMsgServices; # Get defined services..

  my ( $sLine, $sSwitch, $i, $sKey, $bEnabled, $sSmtpPidFile,
       $sImtaCnf, $sResourceName, $sResourceType, $sPIDDir,
       @lsErrServs, %hsServices, %hsMsgServices ) = ();

  #---------------------------------------------------------
  # Populate the data for each service based on ServerType..
  #---------------------------------------------------------
  $sSwitch = $sServerType =~ /^(M(SG|MP))$/o ? $sServerType : 'DEFAULT';
  
  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20078, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20079, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  #----------------
  # C-like switch..
  #----------------
  #goto $sSwitch; {
   if ( "MMP" eq $sSwitch ) {
      push @{ $hsMsgServices { 'pop' }   }, qw ( pidfile.mmp AService );
      push @{ $hsMsgServices { 'pops' }  }, qw ( pidfile.mmp AService );
      push @{ $hsMsgServices { 'imap' }  }, qw ( pidfile.mmp AService );
      push @{ $hsMsgServices { 'imaps' } }, qw ( pidfile.mmp AService );
   }
   elsif ( "MSG" eq $sSwitch ) {
      #-----------
      # Services..
      #-----------
      push @{ $hsMsgServices { 'snmp' }  }, qw ( pidfile.snmp madmand );
      push @{ $hsMsgServices { 'sms' }   }, qw ( pidfile.sms  sms_master );
      push @{ $hsMsgServices { 'ens' }   }, qw ( pidfile.ens  enpd );
      push @{ $hsMsgServices { 'pop' }   }, qw ( pidfile.pop  popd );
      push @{ $hsMsgServices { 'pops' }  }, qw ( pidfile.pop  popd );
      push @{ $hsMsgServices { 'http' }  }, qw ( pidfile.http mshttpd );
      push @{ $hsMsgServices { 'https' } }, qw ( pidfile.http mshttpd );
      push @{ $hsMsgServices { 'imap' }  }, qw ( pidfile.imap imapd );
      push @{ $hsMsgServices { 'imaps' } }, qw ( pidfile.imap imapd );
      push @{ $hsMsgServices { 'smtp' }  }, qw ( pidfile.SMTP tcp_smtp_server
                                                 pidfile.SMTP_SUBMIT tcp_smtp_server
                                                 pidfile.imta_dispatch dispatcher
                                                 pidfile.imta_jbc job_controller );

      #------------------------------
      # Decide on PID file for LMTP..
      #------------------------------
      $sImtaCnf = "$sServerRoot/config/imta.cnf";
      $sSmtpPidFile = 'pidfile.LMTP'; # Set to default..

      #--------------------------------------------
      # Try getting lmtp definition from imta.cnf..
      #--------------------------------------------

      if ( -f $sImtaCnf &&                # If imta.cnf exists..
            $$rhsMsgServices { lmtp } ) { # and LMTP is defined.. 
        if ( open ( IMTA, $sImtaCnf ) ) {
          &logTrace ( 20080,
          "<%s> Opened IMTA file [%s] for read", $sMyName, $sImtaCnf );

          #---------------------
          # Decide on PID file..
          #---------------------
          if ( grep { ( s/\n// || 1 ) && /tcp_lmtpss-daemon/o } <IMTA> ) {
            $sSmtpPidFile = 'pidfile.LMTPSS';
          }

          close ( IMTA ) ||
          &logWarn ( 20081,
          "<%s> Close failed on IMTA file [%s] after read",
          $sMyName, $sImtaCnf );
        }
        else {
          &logWarn ( 20082,
          "<%s> Open failed on IMTA file [%s] for read",
          $sMyName, $sImtaCnf );
        }

        &logTrace ( 20083,
        "<%s> PID file for LMTP set to [%s]", $sMyName, $sSmtpPidFile );
      }

      push @{ $hsMsgServices { 'lmtp' } }, $sSmtpPidFile, 'tcp_lmtp_server',
                                           qw ( pidfile.imta_dispatch dispatcher
                                                pidfile.imta_jbc job_controller );

      #-----------
      # The core.. 
      #-----------
      push @{ $hsMsgServices { 'core' } },  qw ( pidfile.store stored
                                                 pidfile.sched imsched
                                                 pidfile.watcher watcher );
   }
   elsif ( "DEFAULT" eq $sSwitch ) {
      &logError ( 20084,
      "<%s> ServerType [%s] has to be either MSG or MMP",
      $sMyName, $sServerType );
      delete @$rhsMsgServices { @lsServices }; # Delete all..
      $iReturnCode = 0;                        # Failure..
   }

  $iReturnCode || return $iReturnCode; # If failed, return..

  #-----------------------
  # Validate MsgServices..
  #-----------------------
  if ( ! scalar @lsServices ) {
    #-------------
    # Emtpy call..
    #-------------
    &logError ( 20085, "<%s> Called with [%s] services",
    $sMyName, scalar @lsServices );
    $iReturnCode = 0; # Failure..
  }
  elsif ( scalar @lsServices > 1 &&
          map { /^core$/ } @lsServices ) {
    #-------------------------
    # Core not all by itself..
    #-------------------------
    @lsErrServs = grep { ! /^core$/ } @lsServices; # Defaulters..

    &logError ( 20086,
    "<%s> Service [core] cannot be accompanied by other services [%s]",
    $sMyName, join ( ' ', @lsErrServs ) );

    delete @$rhsMsgServices { @lsServices };                         # Delete all..
    @$rhsMsgServices { @lsErrServs } = 'ERROR' x scalar @lsErrServs; # Assign..
    $iReturnCode = 0; # Failure..
  }
  elsif ( scalar ( @lsErrServs =
          grep { ! $hsMsgServices { $_ } } @lsServices ) ) {
    #--------------------
    # Validate services..
    #--------------------
    &logError ( 20087,
    "<%s> Called with un-supported services [%s]. Should be one of [%s]",
    $sMyName, join ( ' ', @lsErrServs ), join ( ' ', sort keys %hsMsgServices ) );
   
    delete @$rhsMsgServices { @lsServices };                         # Delete all..
    @$rhsMsgServices { @lsErrServs } = 'ERROR' x scalar @lsErrServs; # Assign..
    $iReturnCode = 0; # Failure..
  }
  elsif ( scalar ( @lsErrServs =
          grep { $$rhsMsgServices { $_ } =~ /\D/ || # Non-digits..
                 $$rhsMsgServices { $_ } < 1        # Less than 1..
          } @lsServices ) ) {
    #--------------------
    # Validate Port Nos..
    #--------------------
    &logError ( 20088,
    "<%s> Service(s) [%s] do not have port no. correctly set",
    $sMyName, join ( ' ', @lsErrServs ) );

    delete @$rhsMsgServices { @lsServices };                         # Delete all..
    @$rhsMsgServices { @lsErrServs } = 'ERROR' x scalar @lsErrServs; # Assign..
    $iReturnCode = 0; # Failure..
  }
  else {
    #----------------------------------------
    # Update each element with correct path..
    #----------------------------------------
    foreach $sKey ( @lsServices ) {
      #----------------------
      # Even elements first..
      #----------------------
      for ( $i = 0; $i < scalar @{ $hsMsgServices { $sKey } }; $i += 2 ) {
        ${ $hsMsgServices { $sKey } }[$i] =
        "$sPIDDir/" . ${ $hsMsgServices { $sKey } }[$i];
      }

      #------------------------
      # Then the odd elements..
      #------------------------
      for ( $i = 1; $i < scalar @{ $hsMsgServices { $sKey } }; $i += 2 ) {
        ${ $hsMsgServices { $sKey } }[$i] =
        "$sServerRoot/lib/" . ${ $hsMsgServices { $sKey } }[$i];
      }
    }

    delete @$rhsMsgServices { @lsServices };                           # Delete all..
    @$rhsMsgServices { @lsServices } = @hsMsgServices { @lsServices }; # Assign..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String LDAPTestUser,
                          IN Integer LDAPTestPasswd,
                          IN String MsgHost,
                          IN RefHash MsgServices,
                          IN String MonitorProgram,
                          IN String ResourceName,
                          IN String ServerRoot,
                          IN String ServerType,
                          IN Integer SecondLevelMonitor );

 Description : Performs the in-depth monitoring

 Parameters :

 $sLDAPTestPasswd           Encrypted password for test LDAP user
 $sLDAPTestUser             Test user on LDAP for monitoring
 $sMsgHost                  Hostname/IPv4 address of virtual host for Messaging Server
 $rhMsgServices             Services being hosted by this Messaging Server
 $sMonitorProgram           Full path to external monitor program
 $sResourceName             Name of this VCS resource
 $sServerRoot               Absolute path to server root of Messaging Server
 $sServerType               Role being played by the Messaging Server
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sLDAPTestUser,
       $sLDAPTestPasswd,
       $sMsgHost,
       $rhMsgServices,
       $sMonitorProgram,
       $sResourceName,
       $sServerRoot,
       $sServerType,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20089, "Subroutine <%s> called:
  LDAPTestUser           [%s]
  LDAPTestPasswd         [%s]
  MsgHost                [%s]
  rhMsgServices          [%s]
  MonitorProgram         [%s]
  ResourceName           [%s]
  ServerRoot             [%s]
  ServerType             [%s]
  SecondLevelMonitor     [%s]",

  $sMyName,
  $sLDAPTestUser,   $sLDAPTestPasswd,  $sMsgHost,   $rhMsgServices,
  $sMonitorProgram,   $sResourceName,  $sServerRoot,  $sServerType,
  $iSecondLevelMonitor );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #-------------------------------------------------------
  # Upper bound..
  # Time in EPOCH when entry pt started + VCS time limit..
  #-------------------------------------------------------
  my $iTillWhen = $^T + SunJESMsg::GetEngineTimeout ( 1 ); # With margin..

  #----------------------------
  # Get current resource type..
  #----------------------------
  my $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceType = 'SunJESMsg' if $sResourceType !~ /\w/;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor,
                                           $sResourceType,
                                           $sResourceName ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SunJESMsg::DelayedAttributeCheck ( $sServerRoot,
                                            $sMsgHost,
                                            $sLDAPTestUser,
                                            $sLDAPTestPasswd ) ) ) {
      &logTrace ( 20090,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20091,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20092,
    "<%s> Proceeding with second level monitoring", $sMyName );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SunJESMsg::SecondLevelMonitor ( $sLDAPTestPasswd,
                                         $sLDAPTestUser,
                                         $sMsgHost,
                                         $rhMsgServices,
                                         $sServerRoot,
                                         $sServerType,
                                         $iTillWhen ) ) ) {
    #----------------------------------
    # Online after in-depth monitoring..
    #----------------------------------
      &logTrace ( 20093,
      "<%s> Messaging Server online with status [%s] " .
      "after second level monitoring",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20094,
      "<%s> Messaging Server NOT online. Status [%s] " .
      "after second level monitoring",
      $sMyName, $iReturnCode );
      return $iReturnCode; # Get out..
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..
  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                      # Defined..
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =  # Check status..
         SunJESMsg::ExternalMonitor ( $sExtCmd,          # After execution..
                                      $sExtCmdArgs,
                                      $iTillWhen ) ) ) {

      &logTrace ( 20095,
      "<%s> Messaging server found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20096,
      "<%s> Messaging Server NOT online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : Int DelayedAttributeCheck ( $sServerRoot,
                                        $sMsgHost,
                                        $sLDAPTestUser,
                                        $sLDAPTestPasswd );

 Description : Just In Time Validation

 Parameters :

 $sServerRoot       Absolute path to server root of Messaging Server
 $sMsgHost          Hostname/IPv4 address of virtual host for Messaging Server
 $sLDAPTestUser     Test user on LDAP for monitoring
 $sLDAPTestPasswd   Encrypted password for test LDAP user

 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerRoot,
       $sMsgHost,
       $sLDAPTestUser,
       $sLDAPTestPasswd ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20097, "Subroutine <%s> called:
  ServerRoot       [%s]
  MsgHost          [%s]
  LDAPTestUser     [%s]
  LDAPTestPasswd   [%s]",

  $sMyName,
  $sServerRoot, $sMsgHost, $sLDAPTestUser, $sLDAPTestPasswd );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( ServerRoot MsgHost LDAPTestUser LDAPTestPasswd );
  my @lsReqdArgs = qw ( ServerRoot MsgHost );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20098,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #-----------------------------------------
  # Validate each of the attribute in turn..
  #-----------------------------------------

  #-------------------------
  # Server root must exist..
  #-------------------------
  if ( ! -d $sServerRoot ) {
    &logError ( 20099,
    "<%s> ServerRoot directory [%s] does not exist",
    $sMyName, $sServerRoot );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Validate hostname..
  #--------------------
  if ( ! inet_aton ( $sMsgHost ) ) {
    &logError ( 20100, "<%s> Could not resolve MsgHost [%s]",
    $sMyName, $sMsgHost );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-----------------
  # LDAPTestPasswd..
  #-----------------
  if ( $sLDAPTestUser =~ /\S/ &&    # Defined..
       $sLDAPTestPasswd !~ /\S/ ) { # And not defined..
    &logError ( 20101,
    "<%s> LDAPTestPasswd cannot be NULL [%s] if LDAPTestUser [%s] is defined",
    $sMyName, $sLDAPTestPasswd, $sLDAPTestUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String LDAPTestUser,
                                     IN Integer LDAPTestPasswd,
                                     IN String MsgHost,
                                     IN RefHash MsgServices,
                                     IN String ServerRoot,
                                     IN String ServerType,
                                     IN Integer TillWhen );

 Description : Performs the in-depth monitoring

 Notes:

 1. SecondLevelMonitor is not supported for  snmp  mainly  because
 snmp  is  provided in Msg Server for monitoring via a UDP socket,
 which the agent would block.

 2. SecondLevelMonitor is not supported for ens,  because  it  was
 observed that it caused the Msg server to shut it down all by its
 own, thereby faulting the resource.

 Parameters :

 $sLDAPTestPasswd   Encrypted password for test LDAP user
 $sLDAPTestUser     Test user on LDAP for monitoring
 $sMsgHost          Hostname/IPv4 address of virtual host for Msg Server
 $rhMsgServices     Services being hosted by this Messaging Server
 $sServerRoot       Absolute path to server root of Messaging Server
 $sServerType       Role being played by the Messaging Server
 $iTillWhen         Absolute time in EPOCH to engine timeout

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sLDAPTestPasswd,
       $sLDAPTestUser,
       $sMsgHost,
       $rhMsgServices,
       $sServerRoot,
       $sServerType,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20102, "Subroutine <%s> called:
  LDAPTestPasswd       [%s]
  LDAPTestUser         [%s]
  MsgHost              [%s]
  MsgServices          [%s]
  ServerRoot           [%s]
  ServerType           [%s]
  TillWhen             [%s]",

  $sMyName,
  $sLDAPTestPasswd, $sLDAPTestUser, $sMsgHost, $rhMsgServices,
  $sServerRoot, $sServerType, $iTillWhen );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sMonitorCmd = "$sServerRoot/sbin/immonitor-access";
  my $sDecryptCmd = $sCmdVcsDecrypt; # Get global..

  #-------------------
  # Define CHIPS!.. :)
  #-------------------
  my %hsImmHash = ( 'lmtp' => "-C $sMsgHost",
                    'http' => "-H $sMsgHost",
                    'imap' => "-I $sMsgHost",
                    'pop'  => "-P $sMsgHost",
                    'smtp' => "-S $sMsgHost" );
                    

  my ( $sLine, $sCmd, $sCmdArgs, $sEnc, $sKey, $sTmp, $sDir, $sDomain,
       $sSmtp, $iTimeout, $iExitCode, $iPort, $bUseImm, @lsSockChk,
       @lsCmdOut, @lsFiles ) = ();

  #----------------------------------------------------------
  # Check which processes can be used with immonitor-access..
  #----------------------------------------------------------
  foreach $sKey ( keys %$rhMsgServices ) {
    if ( $hsImmHash { $sKey } &&
         $sLDAPTestUser =~ /\S/ ) { # A test LDAP user exists..

      #-------------------------
      # Handle SMTP separately..
      #-------------------------
      if ( $sKey =~ /(s|l)mtp/o ) {
        if ( ( $sDomain ) = $sMsgHost =~ /\.(.*)/ ) { # Extract domain..
          $sSmtp = "-u $sLDAPTestUser\@$sDomain " .
                   "$hsImmHash{$sKey}:$$rhMsgServices{$sKey}=THRESH";

          &logTrace ( 20103, "<%s> Extracted domain [%s] from FQDN [%s]",
          $sMyName, $sDomain, $sMsgHost );
        }
        else {
          &logError ( 20104,
          "<%s> Attribute MsgHost [%s] is not a Fully Qualified Domain Name",
          $sMyName, $sMsgHost );

          return VCSDef_MonitorExit::UNKNOWN; # Error..
        }
      }
      else { 
        $sTmp .= "$hsImmHash{$sKey}:$$rhMsgServices{$sKey}=THRESH "; 
      }

      ++ $bUseImm; # Set flag..
      &logTrace ( 20105,
      "<%s> Will check service [%s] port [%s] via immonitor-access",
      $sMyName, $sKey, $$rhMsgServices { $sKey } );
    }
    elsif ( $sKey eq 'core' ) { # Handle core..
      &logTrace ( 20106,
      "<%s> Got request to monitor watcher at port [%s]",
      $sMyName, $$rhMsgServices { $sKey } );
      @lsSockChk = $sKey; # Assign..
      last; # Get out..
    }
    elsif ( $sKey eq 'snmp' || # Skip snmp..
            $sKey eq 'ens' ) { # Skip ens..
      &logTrace ( 20107,
      "<%s> SecondLevelMonitor not supported. Skipping [%s] on port [%s]",
      $sMyName, $sKey, $$rhMsgServices { $sKey } );
    }
    else {
      &logTrace ( 20108,
      "<%s> Will check service [%s] port [%s] via socket",
      $sMyName, $sKey, $$rhMsgServices { $sKey } );

      push ( @lsSockChk, $sKey ); 
    }
  }

  #----------------------------------
  # Handle socket based check first..
  #----------------------------------
  foreach $sKey ( @lsSockChk ) {
    $iPort = $$rhMsgServices { $sKey }; # Get Port..
    #--------------------
    # Get time bearings..
    #--------------------
    if ( ( $iTimeout = $iTillWhen - ( time + 5 ) ) <= 10 ) {
      &logError ( 20109,
      "<%s> Insufficient time to do in-depth monitoring. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
      return VCSDef_MonitorExit::UNKNOWN; # Error..
    }

    if ( $sKey eq 'core' ) {
      #--------------------
      # Monitor the store..
      #--------------------
      $sCmd = "$sServerRoot/lib/stored";
      $sCmdArgs = "-tv";

      #-------------------
      # Validate command..
      #-------------------
      if ( ! -e $sCmd ) {
        &logError ( 20110,
        "<%s> Store monitor utility [%s] does not exist",
        $sMyName, $sCmd );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Error..
      }
      elsif ( ! -x $sCmd ) {
        &logError ( 20111,
        "<%s> Store monitor utility [%s] not executable",
        $sMyName, $sCmd );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Error..
      }

      unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
        return $iReturnCode;
      }

      &logTrace ( 20112,
      "<%s> Checking store daemon with Cmd [%s] Args [%s] within [%s] sec",
      $sMyName, $sCmd, $sCmdArgs, $iTimeout );

      #---------------
      # Fire command..
      #---------------
      SunJESMsg::ExecLocalCmd ( $sCmd,
                                $sCmdArgs,
                                \$iExitCode,
                                $iTimeout,
                                \@lsCmdOut );

      #-------------------------------------------------------
      # Exit code: 1 Store is up. Exit code: 0 Store is down..
      #-------------------------------------------------------
      if ( $iExitCode ) { # Store is up..
        #-------------------------------------------
        # Check sleepycat transaction checkpointing..
        #-------------------------------------------
        $sDir = "$sServerRoot/data/store/mboxlist";
 
        if ( -d $sDir ) {
         if ( opendir ( MBOX, $sDir ) ) {
           #---------------------
           # Get log file count..
           #---------------------
           if ( scalar ( @lsFiles = grep { /^log\.\d+$/ }
                                     readdir ( MBOX ) )  > 3 ) {
             &logWarn ( 20113,
             "<%s> Located [%s] sleepycat transaction checkpointing log files [%s].
             Database checkpointing might not be happening, or possible problems
             with stored", $sMyName, scalar @lsFiles, join ( ' ', @lsFiles ) );
           }
           else {
             &logTrace ( 20114,
             "<%s> Located [%s] sleepycat transaction checkpointing log files",
             $sMyName, join ( ' ', @lsFiles ) );
           }

           closedir ( MBOX ) ||
           &logWarn ( 20115,
           "<%s> Could not close mboxlist directory [%s] after read: [%s]",
           $sMyName, $sDir, $! );
         }
         else {
           &logWarn ( 20116,
           "<%s> Could not open mboxlist directory [%s] for read: [%s]",
           $sMyName, $sDir, $! );
         }
        }
        else {
          &logTrace ( 20117, "<%s> Could not locate mboxlist directory [%s]",
          $sMyName, $sDir );
        }
      }
      elsif ( ! defined $iExitCode ) {
        $lsCmdOut [0] ||= 'undef'; # Set if not defined..

        if ( $lsCmdOut [0] eq 'timeout' ) {
          $iReturnCode = VCSDef_MonitorExit::TIMEOUT;

          &logError ( 20118,
          "<%s> Command [%s] with args [%s] timed out. Returning [%s]",
          $sMyName, $sCmd, $sCmdArgs, $iReturnCode );
        }
        else {
          $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
        }
      }
      else {
        $iReturnCode = VCSDef_MonitorExit::OFFLINE;
      }

      &logTrace ( 20119, "<%s> Store status [%s]",
      $sMyName, $lsCmdOut [0] );

      &logTrace ( 20120,
      "<%s> Checking watcher on Host [%s] at Port [%s]",
      $sMyName, $sMsgHost, $iPort );
    }
    else {
      &logTrace ( 20121,
      "<%s> Checking service [%s] on Host [%s] at Port [%s]",
      $sMyName, $sKey, $sMsgHost, $iPort );
    }

    #-----------------
    # Check the port..
    #-----------------
    if ( Comms::CheckPort ( $sMsgHost,
                            $iPort ) ) {
      &logTrace ( 20122,
      "<%s> Service [%s] listening on Host [%s] at Port [%s]",
      $sMyName, $sKey, $sMsgHost, $iPort );
    }
    else {
      &logError ( 20123,
      "<%s> Service [%s] NOT listening on Host [%s] at Port [%s]",
      $sMyName, $sKey eq 'core' ? 'watcher': $sKey, $sMsgHost, $iPort );

      #-----------------------
      # Avoid watcher status..
      #-----------------------
      if ( $sKey ne 'core' ) {
        $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
      }
    }
  }

  #--------------------------------------
  # Check if test user has been defined..
  #--------------------------------------
  if ( $bUseImm ) { # Flag has been set..

    #--------------------------------------
    # Check if vcsdecrypt(1m) can be used..
    #--------------------------------------
    if ( ! -e $sDecryptCmd ) {
      &logError ( 20124, "<%s> Decrypt utility [%s] does not exist",
      $sMyName, $sDecryptCmd );
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    elsif ( ! -x $sDecryptCmd ) {
      &logError ( 20125, "<%s> Decrypt utility [%s] not executable",
      $sMyName, $sDecryptCmd );
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    else {
      &logTrace ( 20126, "<%s> Decrypt utility [%s] located",
      $sMyName, $sDecryptCmd );
    }

    #----------------------------------------
    # Check if immonitor-access can be used..
    #----------------------------------------
    if ( ! -e $sMonitorCmd ) {
      &logError ( 20127, "<%s> Monitor utility [%s] does not exist",
      $sMyName, $sMonitorCmd );
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    elsif ( ! -x $sMonitorCmd ) {
      &logError ( 20128, "<%s> Monitor utility [%s] not executable",
      $sMyName, $sMonitorCmd );
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    else {
      &logTrace ( 20129, "<%s> Monitor utility [%s] located",
      $sMyName, $sMonitorCmd );
    }

    #----------------------------------
    # Exit, if not errors encountered..
    #----------------------------------
    unless ( $iReturnCode == VCSDef_MonitorExit::ONLINE ) {
      return $iReturnCode;
    }

    #--------------------
    # Get time bearings..
    #--------------------
    if ( ( $iTimeout = $iTillWhen - ( time + 5 ) ) <= 10 ) {
      &logError ( 20130,
      "<%s> Insufficient time to do in-depth monitoring. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
      return VCSDef_MonitorExit::UNKNOWN; # Error..
    }

    #------------------------
    # Form command and args..
    #------------------------
    if ( $sSmtp ) { # Handle smtp separately..
      $sCmd = $sMonitorCmd;
      ( $sCmdArgs = $sSmtp ) =~ s/THRESH/$iTimeout - 5/e;
      ( $iExitCode, @lsCmdOut ) = (); # Reset..

      if ( SunJESMsg::ExecLocalCmd ( $sCmd,
                                     $sCmdArgs,
                                     \$iExitCode,
                                     $iTimeout,
                                     \@lsCmdOut ) ) {
  
        &logTrace ( 20131,
        "<%s> Cmd [%s] with args [%s] executed successfully with exit code [%s]",
        $sMyName, $sCmd, $sCmdArgs, $iExitCode );
      }
      elsif ( ! defined $iExitCode ) { # Timed out..
        $lsCmdOut [0] ||= 'undef'; # Set if not defined..

        if ( $lsCmdOut [0] eq 'timeout' ) {
          $iReturnCode = VCSDef_MonitorExit::UNKNOWN;

          &logError ( 20132,
          "<%s> Command [%s] with args [%s] timed out. Returning [%s]",
          $sMyName, $sCmd, $sCmdArgs, $iReturnCode );
        }
        else {
          $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
        }
      }
      elsif ( 1 == $iExitCode ) { # Threshold exceeded..
        $iReturnCode = VCSDef_MonitorExit::TIMEOUT; # Failure..

        &logError ( 20133,
        "<%s> Threshold limit exceeded for Cmd [%s] with args [%s].
         Exit code [%s]. Cmd output [%s]. Returning [%s]",
        $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ),
        $iReturnCode );
      }
      elsif ( 64 == $iExitCode ) { # Usage errors..
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;

        &logWarn ( 20134,
        "<%s> Cmd [%s] was incorrectly used. Exit code [%s]. Returning [%s]",
        $sMyName, $sMonitorCmd, $iExitCode, $iReturnCode );
      }
      else {
        &logError ( 20135,
        "<%s> Cmd [%s] with args [%s] exited with exit code [%s]. Cmd output [%s]",
        $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
        $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
      }
    }

    #--------------------------------------------
    # Exit, if failed here or only smtp defined..
    #--------------------------------------------
    if ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ||
         ! $sTmp ) {
      return $iReturnCode;
    }

    #--------------------
    # Get time bearings..
    #--------------------
    if ( ( $iTimeout = $iTillWhen - ( time + 5 ) ) <= 10 ) {
      &logError ( 20136,
      "<%s> Insufficient time to do in-depth monitoring. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
      return VCSDef_MonitorExit::UNKNOWN; # Error..
    }

    $sCmd = $sDecryptCmd; # Assign..
    $sCmdArgs = "$sLDAPTestPasswd | $sMonitorCmd -u $sLDAPTestUser -w - $sTmp";
    $sCmdArgs =~ s/THRESH/$iTimeout - 5/ge;
    ( $iExitCode, @lsCmdOut ) = (); # Reset..

    &logTrace ( 20137,
    "<%s> Executing Cmd [%s] with Args [%s] within [%s] sec",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout ); 

    if ( SunJESMsg::ExecLocalCmd ( $sCmd,
                                   $sCmdArgs,
                                   \$iExitCode,
                                   $iTimeout,
                                   \@lsCmdOut ) ) {

      &logTrace ( 20138,
      "<%s> Cmd [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
    elsif ( ! defined $iExitCode ) {
      $lsCmdOut [0] ||= 'undef'; # Set if not defined..

      if ( $lsCmdOut [0] eq 'timeout' ) {
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;

        &logError ( 20139,
        "<%s> Command [%s] with args [%s] timed out. Returning [%s]",
        $sMyName, $sCmd, $sCmdArgs, $iReturnCode );
      }
      else {
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure.. 
      }
    }
    elsif ( 1 == $iExitCode ) { # Threshold exceeded..
      $iReturnCode = VCSDef_MonitorExit::TIMEOUT; # Failure..

      &logError ( 20140,
      "<%s> Threshold limit exceeded for Cmd [%s] with args [%s].
       Exit code [%s]. Cmd output [%s]. Returning [%s]",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ),
      $iReturnCode );
    }
    elsif ( 64 == $iExitCode ) { # Usage errors..
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;

      &logWarn ( 20141,
      "<%s> Cmd [%s] was incorrectly used. Exit code [%s]. Returning [%s]",
      $sMyName, $sMonitorCmd, $iExitCode, $iReturnCode );
    }
    else {
      &logError ( 20142,
      "<%s> Cmd [%s] with args [%s] exited with exit code [%s]. Cmd output [%s]",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }

    #------------------------------------------------------------------
    # If the LDAP were down when the Msg server got started  the  cache
    # might not have the information needed to log the test user on. In
    # that case the exit code would be non-zero  for  failure.  On  the
    # other hand, if the cache contained the reqd information, the exit
    # code would be zero, even when LDAP was down.
    #
    # Also if the ServerType  was  'MMP',  and  the  corresponding  Msg
    # Server service was down, the exit code would be non-zero.
    #
    # For the above scenarios, we wish to return UNKNOWN.
    #------------------------------------------------------------------

    #---------------
    # Parse output..
    #---------------
    foreach $sLine ( @lsCmdOut ) {
      if ( $sLine =~ /using cached configuration information/o ) {
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown status..

        &logWarn ( 20143,
        "<%s> Cannot connect to LDAP server. Returning [%s]",
        $sMyName, $iReturnCode );
      }
      elsif ( $sLine =~ /.*\]\s*(.*)/o ) {
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown status..

        &logWarn ( 20144,
        "<%s> Messaging server is facing problems [%s]
        Returning [%s]", $sMyName, $1, $iReturnCode );
      }
      elsif ( $sLine =~ /Internal authentication error/o ||
              $sLine =~ /broken connection while Waiting for response to/ ) {
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown status..

        &logWarn ( 20145,
        "<%s> Problems authenticating LDAPTestUser [%s]. Returning [%s]",
        $sMyName, $sLDAPTestUser, $iReturnCode );
      }
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20146, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  $sResourceType = 'SunJESMsg' if $sResourceType !~ /\w/; # Ensure..
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20147,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
  }
  else {
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    $sCommand = $sCmdHares; # Get global..
    $sCommandArgs = "-value $sResourceName $sAttribute";
 
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                      =~ s/\n//;
 
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20148,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s].",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iEngineTimeout ) { # Undefined value..
      &logTrace ( 20149, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20150, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }

    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    unless ( $iEngineTimeout ) {
      $sCommand = $sCmdHatype; # Get global..
      $sCommandArgs = "-value $sResourceType $sAttribute";
    
      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
    
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20151,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
        $sMyName, $sCmdHatype, $sCommandArgs, $iExitCode );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      elsif ( ! $iEngineTimeout ) { # Empty value..
        &logTrace ( 20152, "<%s> Could not get [%s] from engine",
        $sMyName, $sAttribute );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      else { # Success..
        &logTrace ( 20153, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
      }
    }
    else {
     &logTrace ( 20154, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20155,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

=pod

 -------------------------------------------------------------------------
 Function : bool KillMsgProcs ( IN String ServerRoot,
                                IN String ServerType,
                                IN RefHash MsgServices );

 Description : Kills processes for a Sun ONE Messaging Server

 Parameters :

 $sServerRoot       Absolute path to ServerRoot of Messaging Server
 $sServerType       Role being played by the Messaging Server
 $rhMsgServices     Services being hosted by this Messaging Server

 Returns :  boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub KillMsgProcs {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerRoot,
       $sServerType,
       $rhMsgServices ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20156, "Subroutine <%s> called:
  ServerRoot        [%s]
  ServerType        [%s]
  MsgServices       [%s]",
  $sMyName, $sServerRoot, $sServerType, $rhMsgServices );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..
  my $bRepeat = 0;
  my $iSleepTime = 10; # Sec.. # VERIFY..

  #----------------------
  # Set local variables..
  #----------------------
  my ( @liGoners, @liUnDead ) = ();

  #--------------------------
  # If called during online..
  #--------------------------
  if ( ( caller 1 )[3] eq 'SunJESMsg::Start' ) {
    @liGoners = @lsRunningProcs; # Get global..
  }
  else {
    #----------------------------
    # Get processes for killing..
    #----------------------------
    SunJESMsg::GetProcState ( $sServerRoot,
                              $sServerType,
                              \@liGoners,
                              $rhMsgServices,
                              0 ); # Regular..
  }

  do {
    #----------------------------
    # Get processes for killing..
    #----------------------------
    $bRepeat && SunJESMsg::GetProcState ( $sServerRoot,
                                          $sServerType,
                                          \@liGoners,
                                          $rhMsgServices,
                                          1 ); # Be fast..
  
    #------------------------------------------------------------------
    # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
    # valid process before we actually issue the SIGKILL. We don't want
    # to report a failure for SIGKILL in case  of  defunct  or  missing
    # processes.
    #------------------------------------------------------------------
    if ( scalar @liGoners ) {
      #----------------------
      # Failed to kill some..
      #----------------------
      if ( @liUnDead = grep {
                           kill ( 0, $_ ) && # Valid..
                         ! kill ( 9, $_ )    # and could NOT be killed..
                       } @liGoners ) {       # From what's found..
        &logError ( 20157,
        "<%s> PIDs not killed [%s] Total not killed [%s]. Attempt [%s]",
        $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead, $bRepeat + 1 );
        $iReturnCode = 0; # Failure..
      }
      else { # All killed..
        &logInfo ( 20158,
        "<%s> Killed PIDs [%s] Total killed: [%s], Attempt [%s]",
        $sMyName, join ( ' ', @liGoners ), scalar @liGoners, $bRepeat + 1 );
        $iReturnCode = 1; # Re-set..
      }

      ++ $bRepeat;                    # Attempt again..
      sleep $iSleepTime;              # After sometime..
    }
    else {
      &logTrace ( 20159,
      "<%s> Could not locate any process for ServerRoot [%s], Attempt [%s]",
      $sMyName, $sServerRoot, $bRepeat + 1 );
      $bRepeat = 2; # Cannot 'last'.. so get out thus..
    }
  } while ( $bRepeat < 2 ); # Twice max..

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sCmd        Absolute path to external monitor program
 $sCmdArgs    Command line arguments, if any, for Cmd
 $iTillWhen   Represents the upper limit set by engine in EPOCH 

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   TIMEOUT; State couldn't be determined    VCSDef_MonitorExit::TIMEOUT
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20160, "Subroutine <%s> called with:
  Cmd         [%s]
  CmdArgs     [%s]
  TillWhen    [%s]",

  $sMyName,
  $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdOut ) = ();

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20161,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20162,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + 5 ) ) <= 5 ) { # Lack of time..
   &logError ( 20163,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;  
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20164,
  "<%s> Executing external monitor with Command [%s] Args " .
  "[%s] timeout [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  SunJESMsg::ExecLocalCmd ( $sCmd,
                            $sCmdArgs,
                            \$iExitCode,
                            $iTimeout,
                            \@lsCmdOut );

  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..

    if ( $lsCmdOut [0] eq 'timeout' ) {
      $iExitCode = VCSDef_MonitorExit::TIMEOUT;

      &logError ( 20165,
      "<%s> Command [%s] with args [%s] timed out. Returning [%s]",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
    else {
      $iExitCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20166,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }
  else {
    &logWarn ( 20167,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20168, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20169, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20170,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20171,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20172,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20173, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20174,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  #goto $sSwitch; {
  if ( "solaris" eq $sSwitch ) {
      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  }
  elsif ( "hpux" eq $sSwitch ) {
      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }
  elsif ( "aix" eq $sSwitch ) {
      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  }
  elsif ( "linux" eq $sSwitch ) {
      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }

  &logTrace ( 20175, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20176,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20177,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference CmdOut );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdOut        Stores o/p from command if successful

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdOut ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20178, "Subroutine <%s> called with:
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdOut   [%s]",

  $sMyName,
  $sCommand, $sCommandArgs, $riExitCode, $iTimeout, $rlsCmdOut );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-----------------------------------
  # Verify timeout is greater than 1..
  #-----------------------------------
  if ( $iTimeout <= 1 ) {
    &logError ( 20179,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    return 0;
  }

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0;   # Do not source..
  my $sEnvFile = "";        # No env file..
  my $sUser = 'root';       # Run as root..
  my $sTmpFile = tmpnam (); # Outfile..
  my $bTimedOut = 0;        # Timeout variable..

  @$rlsCmdOut = (); # Reset..

  $sCommandArgs .= " 1> $sTmpFile"; # Append..

  &logTrace ( 20180,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdOut,
                                                           \$bTimedOut,
                                                           $iTimeout );
  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20181,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20182,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }

    #-------------------------
    # Now collect the output..
    #-------------------------
    if ( -s $sTmpFile ) {
      if ( open ( LOGFILE, $sTmpFile ) ) {
        @$rlsCmdOut = grep { s/\n// || 1 } <LOGFILE>; # Read file..

        &logTrace ( 20183, "<%s> Command output :----------:\n[%s]",
        $sMyName, join ( chr 012, @$rlsCmdOut ) );

        close ( LOGFILE ) ||
        &logWarn ( 20184, "<%s> Could not close [%s] after read: [%s]",
        $sMyName, $sTmpFile, $! );
      }
      else {
        &logWarn ( 20185, "<%s> Could not open [%s] for logging: [$!]",
        $sMyName, $sTmpFile, $! );
      }
    }
    else {
      &logTrace ( 20186,
      "<%s> Could not collect output of command", $sMyName );
    }
  }
  else { # Execution error..
    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20187,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdOut [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20188,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
      @$rlsCmdOut = (); # Reset..
    }
  }

  #-----------
  # Clean up..
  #-----------
  unlink ( $sTmpFile ) ||
  &logWarn ( 20189, "<%s> Could not remove [%s] after read: [%s]",
  $sMyName, $sTmpFile, $! );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool SetMsgEnv ( IN String ServerRoot,
                             INOUT RefList LibPath );

 Description : Sets/restores enviroment for/after running msstart.

 Parameters :

 $sServerRoot      Absolute path to server root of Messaging Server
 $rlsLibPath       Ref to List of enviroment variables set/saved      

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub SetMsgEnv {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerRoot,
       $rlsLibPath ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20190, "Subroutine <%s> called with:
  ServerRoot              [%s]
  rlsLibPath              [%s]",

  $sMyName, $sServerRoot, $rlsLibPath );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  return 0 unless -d $sServerRoot; # Validate server root..

  #-----------------------------
  # Initialise local variables..
  #-----------------------------
  my $sSwitch = $^O; # Get current OS..
  my ( @lsLibVars, @lsOldLibPath ) = ();

  #----------------------------------
  # Implement a C look-alike switch..
  #----------------------------------
  #goto $sSwitch; {
  if ( "solaris" eq $sSwitch ) {
      push ( @lsLibVars, 'LD_LIBRARY_PATH' );
  }
  elsif ( "hpux" eq $sSwitch || "aix" eq $sSwitch ) {
      push ( @lsLibVars, 'LD_LIBRARY_PATH',
                         'SHLIB_PATH',
                         'LIBPATH' );
  }
  elsif ( "linux" eq $sSwitch ) {
      push ( @lsLibVars, 'LIBRARY_PATH' );
  }

  #-------------------
  # Save old environ..
  #-------------------
  foreach $sSwitch ( @lsLibVars ) {
    if ( $ENV { $sSwitch } ) {
      push ( @lsOldLibPath, $ENV { $sSwitch } );
      &logTrace ( 20191, "<%s> Env variable [%s] Old value [%s]",
      $sMyName, $sSwitch, $ENV { $sSwitch } );
    }
  }

  #-------------------
  # Set to new value..
  #-------------------
  $ENV { SERVERROOT }  = $sServerRoot;
  $ENV { CONFIGROOT }  = $ENV { SERVERROOT } . "/config";
  $ENV { IMTA_TAILOR } = $ENV { CONFIGROOT } . "/imta_tailor";

  if ( scalar @$rlsLibPath ) {          # If we got any..
    delete @ENV { @lsLibVars };         # First delete all existing..
    @ENV { @lsLibVars } = @$rlsLibPath; # Then restore..
  }
  else { # Set new ones..
    @ENV { @lsLibVars } = "$sServerRoot/lib" x scalar @lsLibVars;
  }

  #-----------------------
  # Display the new ones..
  #-----------------------
  foreach $sSwitch ( @lsLibVars ) {
    &logTrace ( 20192, "<%s> Env variable [%s]. Current value [%s]",
    $sMyName, $sSwitch, $ENV { $sSwitch } );
  }

  @$rlsLibPath = @lsOldLibPath; # Return old values..

  return $iReturnCode;
}

1; # So PERL knows its included..
